/*
 * Perform SPI communication using GPIO operations.
 *
 * Copyright (C) 2018, 2020, 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/spi-gpio.h>
#include <time.h>



Spi_gpio::Spi_gpio(Hw::Gpio_chip *clock_device, int clock_pin,
                   Hw::Gpio_chip *data_device, int data_pin,
                   Hw::Gpio_chip *enable_device, int enable_pin,
                   uint32_t frequency)
: _clock_device(clock_device),
  _clock_pin(clock_pin),
  _data_device(data_device),
  _data_pin(data_pin),
  _enable_device(enable_device),
  _enable_pin(enable_pin),
  _frequency(frequency)
{
  _clock_device->setup(_clock_pin, Hw::Gpio_chip::Output, 1);
  _data_device->setup(_data_pin, Hw::Gpio_chip::Output, 0);
  _enable_device->setup(_enable_pin, Hw::Gpio_chip::Output, 1);
}

/* Send a SPI command. */

void Spi_gpio::send(int bytes, uint8_t data[])
{
  struct timespec ts;
  uint8_t mask;
  int bit, byte;

  if (_frequency)
  {
    ts.tv_sec = 0;
    ts.tv_nsec = 1000000000 / _frequency;
  }

  /* Initialise pin levels. */

  _enable_device->set(_enable_pin, 1);
  _clock_device->set(_clock_pin, 1);
  _data_device->set(_data_pin, 0);

  /* Enter the transmission state. */

  _enable_device->set(_enable_pin, 0);

  /* Clock data using the clock and data outputs. */

  for (byte = 0; byte < bytes; byte++)
  {
    mask = 0x80;

    for (bit = 0; bit < 8; bit++)
    {
      /* NOTE: Data presented on falling clock level and sampled on rising clock
               level. This is SPI mode 3, or 0 given that the enable level is
               driven low immediately before the first bit is presented. */

      _clock_device->set(_clock_pin, 0);
      _data_device->set(_data_pin, data[byte] & mask ? 1 : 0);

      if (_frequency)
        nanosleep(&ts, NULL);

      _clock_device->set(_clock_pin, 1);

      if (_frequency)
        nanosleep(&ts, NULL);

      mask >>= 1;
    }
  }

  _enable_device->set(_enable_pin, 1);
}



/* C language interface. */

void *spi_gpio_get_channel(void *clock_chip, int clock_pin,
                           void *data_chip, int data_pin,
                           void *enable_chip, int enable_pin,
                           uint32_t frequency)
{
  return (void *) new Spi_gpio(reinterpret_cast<Hw::Gpio_chip *>(clock_chip), clock_pin,
                               reinterpret_cast<Hw::Gpio_chip *>(data_chip), data_pin,
                               reinterpret_cast<Hw::Gpio_chip *>(enable_chip), enable_pin,
                               frequency);
}

void spi_gpio_send(void *channel, int bytes, uint8_t data[])
{
  static_cast<Spi_gpio *>(channel)->send(bytes, data);
}
